#include "cfs_mesh_interface.h"

#include <map>
#include <tuple>

namespace NiHu 
{
	
void cfs_mesh_interface::print_debug(std::ostream &os)
{
	std::map<int, std::string> type_map = {
		{0, "Undefined"},
		{1, "Point"},
		{2, "Line 2"},
		{3, "Line 3"},
		{4, "Tria 3"},
		{5, "Tria 6"},
		{6, "Quad 4"},
		{7, "Quad 8"},
		{8, "Quad 9"},
		{9, "Tetra 4"},
		{10, "Tetra 10"},
		{11, "Hexa 8"},
		{12, "Hexa 20"},
		{13, "Hexa 27"},
		{14, "Pyra 5"},
		{15, "Pyra 13"},
		{16, "Wedge 6"},
		{17, "Wedge 15"},
		{18, "Wedge 18"},
		{19, "Polygon"},
		{20, "Polyhedron"},
	};

	std::map<int, unsigned int> counts;

	os << "#Nodes: " << m_coords.rows() << "\n";
	os << "#Elems: " << m_connections.rows() << "\n";
	os << "#Regions: " << m_regions.size() << "\n";

	for (size_t i = 0; i < m_types.size(); ++i)
		counts[m_types[i]]++;
	
	os << "Elements by type:" << "\n";

	for (auto it = type_map.begin(); it != type_map.end(); ++it) {
		auto count_it = counts.find(it->first);
		if (count_it != counts.end()) {
			os << "\t" << it->second << ": " << count_it->second << "\n";
		}
	}

	// Print the regions
	for (auto it = m_regions.begin(); it != m_regions.end(); ++it) {
		os << "Region " << it->first << "\n";
		it->second.print_debug(os, "\t");
	}
}



void cfs_mesh_interface::read(h5_file &file)
{
	static const int MAX_NAME = 256;
	static char name_buff[MAX_NAME];

	m_coords = file.read<coordinates_t>("/Mesh/Nodes/Coordinates");
	m_connections = file.read<connections_t>("/Mesh/Elements/Connectivity");
	m_types = file.read<types_t>("/Mesh/Elements/Types");

	std::vector<std::string> region_names = file.get_subgroup_names("/Mesh/Regions");

	for (auto const &name : region_names)
    {
		m_regions[name] = cfs_mesh_region(file, std::string("/Mesh/Regions/" + name));
	}
}

cfs_mesh_interface::index_pairs_t
cfs_mesh_interface::get_index_pairs(std::string const &s1, std::string const &s2) const
{
	index_pairs_t res;

	cfs_mesh_region const &r1 = m_regions.at(s1);
	cfs_mesh_region const &r2 = m_regions.at(s2);

	cfs_mesh_region::indices_t const &i1 = r1.get_nodes();
	cfs_mesh_region::indices_t const &i2 = r2.get_nodes();

	size_t j = 0;
	for (size_t i = 0; i < i1.size(); ++i)
	{
		for (; j < i2.size(); ++j)
		{
			if (i1[i] == i2[j])
			{
				res.push_back(std::make_pair(i, j));
				break;
			}
		}
	}

	return res;
}

} // of namespace NiHu
