#ifndef NIHU_CFS_MESH_INTERFACE_H_INCLUDED
#define NIHU_CFS_MESH_INTERFACE_H_INCLUDED

#include <Eigen/Eigen>
#include "h5_file.hpp"
#include "cfs_mesh_region.h"

#include "core/mesh.hpp"
#include "library/lib_element.hpp"

#include <iostream>
#include <map>
#include <utility>
#include <vector>

namespace NiHu 
{

class cfs_mesh_interface 
{
	typedef Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor> coordinates_t;
	typedef Eigen::Matrix<int, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor> connections_t;
	typedef Eigen::Matrix<int, Eigen::Dynamic, 1> types_t;
	
	/** \brief Region storage type */
	typedef std::map<std::string, cfs_mesh_region> regions_t;
	typedef std::vector<std::pair<int, int> > index_pairs_t;

public:
	cfs_mesh_interface() = default;

	template<class ...Args>
	mesh<tmp::vector<typename tag2type<Args>::type...> >
	create_mesh(std::string const &region_name, bool is_mirror, Args...);
	
	void read(std::string const &fname);

	void read(h5_file &file);

	void print_debug(std::ostream &os = std::cout);

	index_pairs_t get_index_pairs(std::string const &r1, std::string const &r2) const;

private:

	coordinates_t m_coords;
	connections_t m_connections;
	types_t m_types;
	regions_t m_regions;
};

template <class... Args>
mesh<tmp::vector<typename tag2type<Args>::type...> >
cfs_mesh_interface::create_mesh(std::string const &region_name, bool is_mirror, Args...)
{
	static std::map<int, int> cfs2nihu = {
		{4, NiHu::tria_1_elem::id}, 
		{6, NiHu::quad_1_elem::id},
	};

	// Check if region exists
	auto region_it = m_regions.find(region_name);
	if (region_it == m_regions.end())
		throw std::runtime_error("Region \"" + region_name + "\" not found in mesh");
	
	// Extract the region
	cfs_mesh_region const& region = region_it->second;

	// Get the element indices
	cfs_mesh_region::indices_t elements = region.get_elements();
	for (size_t i = 0; i < elements.rows(); ++i)
		elements[i] -= 1;

	size_t num_elements = elements.rows();

	Eigen::MatrixXi elem_mat(num_elements, m_connections.cols() + 1);
	elem_mat << m_types(elements, Eigen::all), m_connections(elements, Eigen::all);

	Eigen::VectorXi node_inds(m_coords.rows());
	Eigen::VectorXi node_sel(m_coords.rows());
	node_inds.setZero();
	node_sel.setZero();

	// Change elem IDs to NiHu type
	for (size_t i = 0; i < elem_mat.rows(); ++i) {
		elem_mat(i, 0) = cfs2nihu[elem_mat(i, 0)];
		// Correct indexing (start from 0 instead of 1)
		for (size_t j = 1; j < elem_mat.cols(); ++j) {
			elem_mat(i, j) -= 1;
			if (elem_mat(i, j) >= 0)
				node_inds[elem_mat(i,j)] = 1;
		}
	}

	size_t c = 0;
	for (size_t i = 0; i < node_inds.rows(); ++i) {
		if (node_inds[i]) {
			node_inds[i] = c;
			node_sel[c] = i;
			++c;
		}
	}

	std::cout << "Resizing node_sel to " << c << std::endl;

	node_sel.conservativeResize(c, 1);

	std::cout << "Renumbering connectivity " << std::endl;

	// Renumber connectivity by only using actually needed nodes
	for (size_t i = 0; i < elem_mat.rows(); ++i) 
		for (size_t j = 1; j < elem_mat.cols(); ++j)
			if (elem_mat(i,j) >= 0)
				elem_mat(i, j) = node_inds[elem_mat(i,j)];

	std::cout << "Renumbering connectivity done" << std::endl;

	coordinates_t node_mat = m_coords(node_sel, Eigen::all);

	if (is_mirror)
	{
		for (size_t i = 0; i < 3; ++i) 
		{
			Eigen::Matrix3d mirror = Eigen::Matrix3d::Identity();
			mirror(i, i) *= -1;
			coordinates_t new_node_mat = node_mat * mirror;
			Eigen::MatrixXi new_elem_mat = elem_mat;
			int n_elems = elem_mat.rows();
			int n_nodes = node_mat.rows();
			for (size_t e = 0; e < n_elems; ++e)
			{
				switch (elem_mat(e, 0)) {
					case NiHu::tria_1_elem::id:
						new_elem_mat.block(e, 1, 1, 3).rowwise().reverseInPlace();
						new_elem_mat.block(e, 1, 1, 3).array() += n_nodes;
						break;
					case NiHu::quad_1_elem::id:
						new_elem_mat.block(e, 1, 1, 4).rowwise().reverseInPlace();
						new_elem_mat.block(e, 1, 1, 4).array() += n_nodes;
						break;
					default:
						throw std::runtime_error("Invalid elem type found while mirroring");
				}
			}
			
			node_mat.conservativeResize(2*n_nodes, node_mat.cols());
			node_mat.bottomRows(n_nodes) = new_node_mat;

			elem_mat.conservativeResize(2*n_elems, elem_mat.cols());
			elem_mat.bottomRows(n_elems) = new_elem_mat;
		}
	}

	// Construct the mesh
	return mesh<tmp::vector<typename tag2type<Args>::type...> >(node_mat, elem_mat);
}

} // of namespace NiHu


#endif /* NIHU_CFS_MESH_INTERFACE_H_INCLUDED */
