/**
 * @file dom_elem.h
 * @author P. Fiala. P. Rucz
 * @brief Declaration of class ::dom_elem
 * @version 0.1
 * @date 2022-08-17
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#ifndef DOM_ELEM_H_INCLUDED
#define DOM_ELEM_H_INCLUDED

#include "dom/DOMElement.hpp"
#include <util/XMLString.hpp>

#include <string>
#include <vector>

/**
 * @brief class representing a node in the xml DOM
 */
class dom_elem
{
public:
	/**
	 * @brief Construct a new dom elem object
	 * 
	 * @param elem xercesc pointer to the appropriate element
	 */
	dom_elem(xercesc::DOMElement *elem = nullptr)
		: m_dom_element(elem)
	{
	}

	/**
	 * @brief Get the first child with a given tag
	 * @param tag tag of the child
	 * @return the first child element with the specified tag
	 */
	dom_elem get_child(std::string const &tag) const;

	/**
	 * @brief Get all children of type tag
	 * @param tag the requested type
	 * @return vector of dom_elem objects
	 */
	std::vector<dom_elem> get_children(std::string const &tag) const;

	/**
	 * @brief Get the attribute name="attr".
	 * @param name the attribute's name
	 * @return the attribute. Empty string if the attribute does not exist.
	 */
	std::string get_attribute(std::string const &name) const;

	/**
	 * @brief Get the attribute under a given path
	 * @param path the path of nodes separated by "/" characters
	 * @param name the attibute's name
	 * @return the attribute's value, or empty string if the attribute does not exist
	 */
	std::string get_attribute(std::string const &path, std::string const &name) const;

	std::string get_text_content() const
	{
		return xmlstr2stdstr(m_dom_element->getTextContent());
	}

	/**
	 * @brief cast a dom_node to bool
	 * @return true if the dom_node is valid
	 * @return false if the dom_node is invalid (nullptr)
	 */
	operator bool() const { return m_dom_element != nullptr; }

private:
	static std::string xmlstr2stdstr(XMLCh const *xmlstr)
	{
		char *c = xercesc::XMLString::transcode(xmlstr);
		std::string str = c;
		xercesc::XMLString::release(&c);
		return str;
	}

	xercesc::DOMElement *m_dom_element;
};

#endif
