#ifndef NIHU_H5_FILE_HPP_INCLUDED
#define NIHU_H5_FILE_HPP_INCLUDED

#include "H5Cpp.h"
#include "H5Apublic.h"
#include "H5Gpublic.h"
#include "H5Lpublic.h"

#include <Eigen/Eigen>

#include <exception>
#include <fstream>
#include <type_traits>
#include <vector>

namespace NiHu
{

template <typename T>
struct h5_type_traits;

template<>
struct h5_type_traits<int>
{
    enum 
    {
            h5_type_class = H5T_INTEGER,
    };

    static H5::PredType get_h5_pred_type() 
    {
        return H5::PredType::NATIVE_INT;
    }

};

template<>
struct h5_type_traits<double>
{
    enum 
    {
            h5_type_class = H5T_FLOAT,
    };

    static H5::PredType get_h5_pred_type() 
    {
        return H5::PredType::NATIVE_DOUBLE;
    }

    //static const H5::PredType h5_pred_type;
};


class h5_file
{
private:
    template <typename T>
    struct reader;
    
    /**
     * @brief Reader class for Eigen stuff
     */
    template<typename Scalar_, int Rows_, int Cols_, int Options_, int MaxRows_, int MaxCols_>
    struct reader<Eigen::Matrix<Scalar_, Rows_, Cols_, Options_, MaxRows_, MaxCols_> >
    {
        typedef Eigen::Matrix<Scalar_, Rows_, Cols_, Options_, MaxRows_, MaxCols_> matrix_t;
        //static_assert(std::is_base_of<Eigen::MatrixBase<Matrix>, Matrix>::value);
        typedef typename matrix_t::Scalar scalar_t;

        static matrix_t read(h5_file &file, std::string const &path)
        {
            // check existence of path
            if ( ! file.exists(path) )
                throw std::runtime_error("h5_file::read: dataset not found ('"+path+"')");

            // open dataset
            H5::DataSet dataset = file.m_file.openDataSet(path.c_str());

            // check precision
            if ( ! file.correct_precision<scalar_t>(dataset) )
                throw std::runtime_error("h5_file::read: precision inconsistent ('"+path+"')");

            // get shape
            std::vector<size_t> shape = file.shape(path);

            // check rank (rank 1 and rank 2 are accepted)
            if ( shape.size() != 2 && shape.size() != 1 )
                throw std::runtime_error("h5_file::read: dataset has rank " + std::to_string(shape.size()) +  " ('"+path+"')");

            // allocate output
            matrix_t data;
            if (shape.size() == 2)
            {
                if (Rows_ == Eigen::Dynamic && Cols_ == Eigen::Dynamic)
                    data.resize(shape[0], shape[1]);
                else if (Rows_ == Eigen::Dynamic)
                {
                    if (Cols_ != shape[1])
                        throw std::runtime_error("h5_file::read: size mismatch (trying to read " + std::to_string(shape[0]) + " x " + std::to_string(shape[1]) + 
                                                 " elements into matrix of type " + "X" + " x " + std::to_string(Cols_) + ")");
                    else
                        data.resize(shape[0], shape[1]);
                } 
                else if (Cols_ == Eigen::Dynamic)
                {
                    if (Rows_ != shape[0])
                        throw std::runtime_error("h5_file::read: size mismatch (trying to read " + std::to_string(shape[0]) + " x " + std::to_string(shape[1]) + 
                                                 " elements into matrix of type " + std::to_string(Rows_) + " x " + "X" + ")");
                    else
                        data.resize(shape[0], shape[1]);
                } 
                else if (Rows_ != shape[0] || Cols_ != shape[1]) 
                {
                    throw std::runtime_error("h5_file::read: size mismatch (trying to read " + std::to_string(shape[0]) + " x " + std::to_string(shape[1]) +
                        " elements into matrix of type " + std::to_string(Rows_) + " x " + std::to_string(Cols_) + ")");
                }
            }
            else if (shape.size() == 1)
            {
                if (Rows_ == Eigen::Dynamic)
                    data.resize(shape[0], 1);
                else 
                {
                    if (Cols_ == Eigen::Dynamic)
                        data.resize(1, shape[0]);
                    else if (Cols_ * Rows_ != shape[0])
                        throw std::runtime_error("h5_file::read: size mismatch (trying to read " + std::to_string(shape[0]) + 
                            " elements into matrix of type " + std::to_string(Rows_)+ " x " + std::to_string(Cols_) + ")");
                }
            }

            // read data
            dataset.read(data.data(), h5_type_traits<scalar_t>::get_h5_pred_type());
            //dataset.read(data.data(), H5::PredType::NATIVE_INT);

            // return output
            return data;
  
        }
    };

public:
    h5_file(const std::string &name, std::string const &mode = "w", bool autoflush = true);

    template <typename T>
    bool correct_precision(H5::DataSet const &dataset);

    /**
     * @brief Check if a path exists in the file (is dataset or group)
     */
    bool exists(std::string const &path) const;

    H5::H5File &get_H5File() {return m_file;}

    template<typename T>
    T read_attribute(std::string const &path, std::string const &name);
    
    /**
     * @brief Read data from the HDF5 file
     */
    template<typename T>
    T read(std::string const &path);

    std::vector<size_t> shape(std::string const &path);

    std::vector<std::string> get_subgroup_names(std::string const& path);

private:
    H5::H5File  m_file;
    std::string m_fname;
    std::string m_mode;
    bool m_autoflush;
};

inline h5_file::h5_file(std::string const &name, std::string const &mode, bool autoflush)
{
    m_mode = mode;
    m_fname = name;

    // check if the file exists
    if ( m_mode == "r" )
    {
        // - find file
        std::ifstream infile(m_fname);
        // - throw error if file does not exist
        if ( ! infile.good() ) 
            throw std::runtime_error("HDF5pp: file does not exist ('"+name+"')");
    }

    // check if file exists, otherwise set write mode to "w"
    if ( m_mode == "a" or m_mode == "r+" )
    {
        // - find file
        std::ifstream infile(m_fname);
        // - change write mode if file does not exist
        if ( ! infile.good() ) 
            m_mode = "w";
    }

    // open file
    if      ( m_mode == "r"         ) m_file = H5::H5File(m_fname.c_str(), H5F_ACC_RDONLY);
    else if ( m_mode == "w"         ) m_file = H5::H5File(m_fname.c_str(), H5F_ACC_TRUNC );
    else if ( m_mode == "a" or "r+" ) m_file = H5::H5File(m_fname.c_str(), H5F_ACC_RDWR  );
    else throw std::runtime_error("h5_file: unknown mode '"+mode+"'");

    // store flush settings
    m_autoflush = autoflush;
}

template <typename T>
inline bool h5_file::correct_precision(H5::DataSet const &dataset)
{
    // check data-type
  if ( static_cast<int>(dataset.getTypeClass()) != static_cast<int>(h5_type_traits<T>::h5_type_class) ) 
    return false;

  // check the number of bytes
  if ( dataset.getIntType().getSize() != sizeof(T) ) 
    return false;

  return true;
}

inline bool h5_file::exists(std::string const &path) const
{
    // find first "/"
    size_t idx = path.find("/");

    // loop over all groups
    while (true)
    {
        // - terminate if all "/" have been found
        if (idx == std::string::npos)
            break;
        // - create group if needed
        if (idx > 0)
        {
            // -- get group name
            std::string name(path.substr(0, idx));
            // -- create if needed
            if (H5Lexists(m_file.getId(), name.c_str(), H5P_DEFAULT) <= 0)
                return false;
        }
        // - proceed to next "/"
        idx = path.find("/", idx + 1);
    }

    return H5Lexists(m_file.getId(), path.c_str(), H5P_DEFAULT) > 0;
}

template<typename T>
T h5_file::read_attribute(std::string const &path, std::string const &name)
{
    H5::Group group = m_file.openGroup(path);
    H5::Attribute attr = group.openAttribute(name);

    T value;

    H5::DataType type = attr.getDataType();
    attr.read(type, &value);

    return value;
}

template<>
std::string h5_file::read_attribute<std::string>(std::string const &path, std::string const &name);

template<typename T>
T h5_file::read(std::string const &path)
{
    return reader<T>::read(*this, path);
}

inline std::vector<size_t> h5_file::shape(std::string const &path)
{
    // check existence of path
    if (!exists(path))
        throw std::runtime_error("h5_file::shape: dataset not found ('" + path + "')");

    // open data-space
    H5::DataSpace dataspace = m_file.openDataSet(path.c_str()).getSpace();

    // get the size in each direction
    // - read rank (a.k.a number of dimensions)
    int rank = dataspace.getSimpleExtentNdims();
    // - allocate as HDF5-type
    std::vector<hsize_t> dimsf(rank);
    // - allocate as vector
    std::vector<size_t> shape(rank);
    // - read
    dataspace.getSimpleExtentDims(dimsf.data(), NULL);
    // - convert to vector
    for (int i = 0; i < rank; ++i)
        shape[i] = static_cast<size_t>(dimsf[i]);

    return shape;
}


} // of namespace NiHu

#endif /* NIHU_H5_FILE_HPP_INCLUDED */
