#include "acoustic_material.h"
#include "cfs_data_interface.h"
#include "cfs_mesh_interface.h"
#include "h5_file.hpp"
#include "xml_file.h"

#include <chrono>
#include <iostream>
#include <string>
#include <vector>

#include <Eigen/IterativeLinearSolvers>

#include <core/function_space.hpp>
#include <core/integral_operator.hpp>
#include <core/integral_transform.hpp>
#include <core/weighted_residual.hpp>
#include <library/helmholtz_3d.hpp>
#include <library/helmholtz_3d_nearly_singular_integrals.hpp>
#include <library/helmholtz_3d_singular_integrals.hpp>

using namespace boost::math::double_constants;
using namespace std::complex_literals;

template <class M>
void matrix_to_file(std::string const& fname, M const& matrix)
{
   // Write normals to file
   std::ofstream ofs(fname);
   for (size_t i = 0; i < matrix.rows(); ++i) {
      for (size_t j = 0; j < matrix.cols(); ++j)
         ofs << matrix(i, j) << '\t';
      ofs << '\n';
   }
   ofs.close();
}

std::vector<double> read_analysis_freqs(dom_elem const &sequence_step)
{
   dom_elem anal_elem = sequence_step.get_child("analysis").get_child("harmonic");
   int num_freq = std::stoi(anal_elem.get_child("numFreq").get_text_content());
   double start_freq = std::stod(anal_elem.get_child("startFreq").get_text_content());
   double stop_freq = std::stod(anal_elem.get_child("stopFreq").get_text_content());
   std::string sampling = anal_elem.get_child("sampling").get_text_content();
   std::vector<double> anal_freqs;
   if (sampling == "linear")
      for (int i = 0; i < num_freq; ++i)
         anal_freqs.push_back(start_freq + i * (stop_freq - start_freq) / num_freq);
   else
      throw std::runtime_error("Unhandled sampling method: " + sampling);
   return anal_freqs;
}

int main()
{
   xml_file::initialize();

   xml_file config_file("mech_acoustic_FWD-coupling_BEM.xml");

   dom_elem config_root = config_file.get_root();

   dom_elem formats = config_root.get_child("fileFormats");
   std::string mesh_name = formats.get_attribute("input/hdf5", "fileName");
   std::string material_name = formats.get_attribute("materialData", "file");

   std::cout << mesh_name << std::endl;
   std::cout << material_name << std::endl;

   xml_file mat_file(material_name);
   dom_elem mat_root = mat_file.get_root();
   std::vector<dom_elem> material_elems = mat_root.get_children("material");
   std::vector<acoustic_material> materials;
   for (auto const &mat : material_elems)
   {
      if (!mat.get_child("acoustic"))
         continue;
      std::string name = mat.get_attribute("name");
      std::string rho_str = mat.get_child("acoustic").get_child("density").get_child("linear").get_child("real").get_text_content();
      std::string kappaP0_str = mat.get_child("acoustic").get_child("compressionModulus").get_child("linear").get_child("real").get_text_content();
      double rho = std::stod(rho_str);
      double kappaP0 = std::stod(kappaP0_str);
      materials.push_back(acoustic_material(name, rho, kappaP0));
   }
   mat_file.close();

   NiHu::h5_file file = NiHu::h5_file(mesh_name, "r");
   NiHu::cfs_mesh_interface cfs_mesh;
   cfs_mesh.read(file);
   cfs_mesh.print_debug();

   auto surface_mesh = cfs_mesh.create_mesh("IF_SourceRegion", false, NiHu::tria_1_tag(), NiHu::quad_1_tag());

   std::cout << "Surface mesh:" << '\n';
   std::cout << "\t#Elems: " << surface_mesh.get_num_elements() << '\n';
   std::cout << "\t#Points: " << surface_mesh.get_num_points() << '\n';


   Eigen::Matrix<double, Eigen::Dynamic, 3> normals(surface_mesh.get_num_elements(), 3);
   Eigen::SparseMatrix<double> A(surface_mesh.get_num_elements(), surface_mesh.get_num_points());
   {
      std::vector<Eigen::Triplet<double> > triplets;
      auto const &c = NiHu::quad_1_elem::lset_t::domain_t::get_center();
      auto N = NiHu::quad_1_elem::lset_t::eval_shape<0>(c);
      for (auto el_it = surface_mesh.begin<NiHu::quad_1_elem>(); el_it != surface_mesh.end<NiHu::quad_1_elem>(); ++el_it)
      {
         NiHu::quad_1_elem const &e = *el_it;
         auto const &node_idx = e.get_nodes();
         auto el_id = e.get_id()(0,0);
         for (size_t i = 0; i < N.rows(); ++i)
            triplets.push_back(Eigen::Triplet<double>(el_id, node_idx(i), N(i)));
         normals.row(el_id) =  e.get_normal(c).normalized();
      }
      A.setFromTriplets(triplets.begin(), triplets.end());
   }

   // Write normals to file
   matrix_to_file("normals.txt", normals);
   

   std::vector<dom_elem> sequence_steps = config_root.get_children("sequenceStep");
   for (auto const &sequence_step : sequence_steps)
   {
      dom_elem bem_sol_elem = sequence_step.get_child("bemSolution");
      if (!bem_sol_elem)
         continue;

      dom_elem vel_elem = bem_sol_elem.get_child("bcsAndLoads").get_child("velocity");
      std::string bc_reg_name = vel_elem.get_attribute("name");
      dom_elem seq_elem = vel_elem.get_child("sequenceStep");
      size_t multistep_index = std::stoi(seq_elem.get_attribute("index"));
      std::string quantity = seq_elem.get_attribute("quantity", "name");

      NiHu::h5_file data_file = NiHu::h5_file("mech_acoustic_FWD-coupling.cfs", "r");
      std::stringstream ss;

      // read data file source frequencies
      ss.str("");
      ss << "/Results/Mesh/MultiStep_" << multistep_index;
      std::string anal_type = data_file.read_attribute<std::string>(ss.str(), "AnalysisType");
      if (anal_type != "harmonic")
         throw std::runtime_error("Source data analysis type is not harmonic!");
      ss.str("");
      ss << "/Results/Mesh/MultiStep_" << multistep_index << "/ResultDescription/" << quantity << "/StepValues";
      auto src_freqs = data_file.read<Eigen::Matrix<double, Eigen::Dynamic, 1>>(ss.str());

      auto anal_freqs = read_analysis_freqs(sequence_step);

      for (auto freq : anal_freqs)
      {
         double omega = two_pi * freq;
         double speed_of_sound = materials[0].get_speed_of_sound();
         double rho = materials[0].get_rho();
         double wave_number = omega / speed_of_sound;

         std::cout << "frequency: " << freq << " Hz" << std::endl ;
         std::cout << "wave number: " << wave_number << " rad/m" << std::endl;

         auto freq_it = std::find_if(src_freqs.begin(), src_freqs.end(), [freq](double f)
                                     { return std::abs(f - freq) < 1e-1; });
         if (freq_it == src_freqs.end())
            throw std::runtime_error("Did not find frequency " + std::to_string(freq) + " in data file");
         size_t step_idx = freq_it - src_freqs.begin() + 1;

         NiHu::cfs_data_interface<std::complex<double>> cdi;
         ss.str("");
         ss << "/Results/Mesh/MultiStep_" << multistep_index << "/Step_" << step_idx << "/" << quantity;
         cdi.read(data_file, ss.str());

         std::cout << "cdi data:" << std::endl;
         std::cout << cdi.get_data("Source_region").topRows(10) << std::endl;

         Eigen::Matrix<std::complex<double>, Eigen::Dynamic, Eigen::Dynamic> filtered_data(surface_mesh.get_num_points(), 3);
         std::vector<std::string> source_region_names = data_file.get_subgroup_names(ss.str());
         std::vector<int> cntr(surface_mesh.get_num_points(), 0);

         for (auto const &src_reg_name : source_region_names)
         {
            auto indices = cfs_mesh.get_index_pairs(src_reg_name, bc_reg_name);
            for (auto const &i : indices)
            {
               filtered_data.row(i.second) = cdi.get_data(src_reg_name).row(i.first);
               cntr[i.second]++;
            }
         }

         auto badit = std::find_if(cntr.begin(), cntr.end(), [](int i)
                                   { return i == 0; });
         if (badit != cntr.end())
            throw std::runtime_error("No data at idx " + std::to_string(badit - cntr.begin()));
         badit = std::find_if(cntr.begin(), cntr.end(), [](int i)
                              { return i > 1; });
         if (badit != cntr.end())
            throw std::runtime_error("Multiple data at idx " + std::to_string(badit - cntr.begin()));

         Eigen::Matrix<std::complex<double>, Eigen::Dynamic, Eigen::Dynamic> vel_elem = A * filtered_data;
         Eigen::Matrix<std::complex<double>, Eigen::Dynamic, 1> vel_bc = (vel_elem * normals.transpose()).diagonal();
         Eigen::Matrix<std::complex<double>, Eigen::Dynamic, 1> q_bc = -1.i * omega * rho * vel_bc;

         std::cout << "velocity in nodes:" << std::endl;
         std::cout << filtered_data.topRows(10) << std::endl;

         std::cout << "velocity in elems:" << std::endl;
         std::cout << vel_elem.topRows(10) << std::endl;

         std::cout << "normal velocity in elems:" << std::endl;
         std::cout << vel_bc.topRows(10) << std::endl;

         std::cout << "gradient excitation:" << std::endl;
         std::cout << q_bc.topRows(10) << std::endl;

         bool is_mirror = true;

         auto full_surface_mesh = cfs_mesh.create_mesh("IF_SourceRegion", is_mirror, NiHu::tria_1_tag(), NiHu::quad_1_tag());
         if (is_mirror)
            for (int i = 0; i < 3; ++i)
            {
               size_t n = q_bc.rows();
               q_bc.conservativeResize(2*n, q_bc.cols());
               q_bc.bottomRows(n) = q_bc.topRows(n);
            }

         std::cout << full_surface_mesh.get_num_elements() << " elements." << std::endl;
         std::cout << full_surface_mesh.get_num_points() << " points." << std::endl;


#if 0
         // Surface normals of full mesh
         Eigen::Matrix<double, Eigen::Dynamic, 3> full_normals(full_surface_mesh.get_num_elements(), 3);
         auto const &c = NiHu::quad_1_elem::lset_t::domain_t::get_center();
         for (auto el_it = full_surface_mesh.begin<NiHu::quad_1_elem>(); el_it != full_surface_mesh.end<NiHu::quad_1_elem>(); ++el_it)
         {
            NiHu::quad_1_elem const &e = *el_it;
            auto el_id = e.get_id()(0,0);
            full_normals.row(el_id) =  e.get_normal(c).normalized();
         }

         // Write normals to file
         matrix_to_file("full_normals.txt", full_normals);
#endif // 0
   
         // Function spaces, integration
         auto const &fs = NiHu::constant_view(full_surface_mesh);

         size_t n = full_surface_mesh.get_num_elements();
         Eigen::Matrix<std::complex<double>, Eigen::Dynamic, 1> rhs;

         {
            Eigen::Matrix<std::complex<double>, Eigen::Dynamic, Eigen::Dynamic> L(n, n);
            L.setZero();

            NiHu::helmholtz_3d_SLP_kernel<double> g(wave_number);
            auto gop = NiHu::create_integral_operator(g);

            auto tstart = std::chrono::high_resolution_clock::now();
            L << NiHu::dirac(fs) * gop[fs];
            rhs = L * q_bc;
            auto tass = std::chrono::high_resolution_clock::now();
            std::cout << "Rhs time: " << std::chrono::duration_cast<std::chrono::seconds>(tass-tstart).count() << std::endl;
         }

         Eigen::Matrix<std::complex<double>, Eigen::Dynamic, 1> res;
         {
            Eigen::Matrix<std::complex<double>, Eigen::Dynamic, Eigen::Dynamic> M(n, n);
            M.setZero();

            NiHu::helmholtz_3d_DLP_kernel<double> h(wave_number);
            auto hop = NiHu::create_integral_operator(h);
            auto iop = NiHu::identity_integral_operator();

            auto tstart = std::chrono::high_resolution_clock::now();
            M << NiHu::dirac(fs) * hop[fs] + dirac(fs) * (-.5*iop)[fs];
            std::cout << "M:" << std::endl;
            std::cout << M.topLeftCorner(5, 5) << std::endl;
            auto tass = std::chrono::high_resolution_clock::now();
            std::cout << "Assembly time: " << std::chrono::duration_cast<std::chrono::seconds>(tass-tstart).count() << std::endl;

            Eigen::BiCGSTAB<decltype(M), Eigen::IdentityPreconditioner > solver(M);
            res = solver.solve(rhs);

            auto tend = std::chrono::high_resolution_clock::now();
            std::cout << "Solution time: " << std::chrono::duration_cast<std::chrono::seconds>(tend-tass).count() << " s" << std::endl;
         }

         std::cout << "result: " << std::endl;
         std::cout << res.topRows(10) << std::endl;

         matrix_to_file("result.txt", res);
      }

   }

   config_file.close();

   xml_file::terminate();

   return 0;
}
